#include "dxf.hpp"


Mat DXF::to8bitBGR() const {
    if (_imageData.empty()) {
        throw runtime_error("No image data loaded");
    }

    Mat output(_height, _width, CV_8UC3);

    double scale = 255.0 / ((1 << _bitDepth) - 1);
    
    for (int y = 0; y < _height; ++y) {
        const uint16_t* srcRow = _imageData.ptr<uint16_t>(y);
        uint8_t* dstRow = output.ptr<uint8_t>(y);
        
        for (int x = 0; x < _width; ++x) {
            uint8_t value = static_cast<uint8_t>(srcRow[x] * scale + 0.5);
            dstRow[x * 3] = value;     // B
            dstRow[x * 3 + 1] = value; // G
            dstRow[x * 3 + 2] = value; // R
        }
    }

    return output;
}

Mat DXF::to10bitBGR() const {
    if (_imageData.empty()) {
        throw runtime_error("No image data loaded");
    }

    Mat output(_height, _width, CV_16UC3);
    
    double scale = 1023.0 / ((1 << _bitDepth) - 1);
    
    for (int y = 0; y < _height; ++y) {
        const uint16_t* srcRow = _imageData.ptr<uint16_t>(y);
        uint16_t* dstRow = output.ptr<uint16_t>(y);
        
        for (int x = 0; x < _width; ++x) {
            uint16_t value = static_cast<uint16_t>(srcRow[x] * scale + 0.5);
            value = min(value, static_cast<uint16_t>(1023));
            
            uint16_t shiftedValue = value << 6;
            
            dstRow[x * 3] = shiftedValue;     // B
            dstRow[x * 3 + 1] = shiftedValue; // G
            dstRow[x * 3 + 2] = shiftedValue; // R
        }
    }

    return output;
}

void DXF::analyzeFile() {
    ifstream file(_filepath, ios::binary);
    if (!file.is_open()) {
        throw runtime_error("Could not open file: " + _filepath);
    }

    if (!readDXFHeader(file)) {
        analyzeDXFContent(file);
    }

    validateImageProperties();
}

bool DXF::readDXFHeader(ifstream& file) {
    struct DXFHeader {
        uint32_t magic;
        uint32_t width;
        uint32_t height;
        uint8_t bitDepth;
        uint8_t version;
        uint16_t flags;
    };

    DXFHeader header;
    file.read(reinterpret_cast<char*>(&header), sizeof(header));

    if (file.gcount() != sizeof(header)) {
        return false;
    }

    if (header.magic != 0x44584631) {
        return false;
    }

    _width = header.width;
    _height = header.height;
    _bitDepth = header.bitDepth;

    return true;
}

void DXF::analyzeDXFContent(ifstream& file) {
    file.seekg(0, ios::end);
    streamsize fileSize = file.tellg();
    file.seekg(0, ios::beg);

    vector<uint8_t> sample(min(static_cast<streamsize>(1024), fileSize));
    file.read(reinterpret_cast<char*>(sample.data()), sample.size());

    _bitDepth = detectBitDepth(sample);

    calculateDimensions(fileSize);
}

int DXF::detectBitDepth(const vector<uint8_t>& sample) {
    uint16_t maxValue = 0;
    
    for (size_t i = 0; i < sample.size() - 1; i += 2) {
        uint16_t value = (static_cast<uint16_t>(sample[i + 1]) << 8) | sample[i];
        maxValue = max(maxValue, value);
    }

    int bitDepth = 16;
    while (bitDepth > 8 && maxValue < (1 << (bitDepth - 1))) {
        bitDepth--;
    }

    return bitDepth;
}

void DXF::loadImageData() {
    ifstream file(_filepath, ios::binary);
    if (!file.is_open()) {
        throw runtime_error("Could not open file: " + _filepath);
    }
    if (hasDXFHeader()) {
        file.seekg(sizeof(struct DXFHeader));
    }

    size_t dataSize = static_cast<size_t>(ceil(_width * _height * _bitDepth / 8.0));
    vector<uint8_t> rawData(dataSize);

    file.read(reinterpret_cast<char*>(rawData.data()), dataSize);

    _imageData = Mat(_height, _width, CV_16UC1);

    unpackBits(rawData);
}

void DXF::unpackBits(const vector<uint8_t>& rawData) {
    size_t bitPos = 0;
    const uint16_t maxValue = (1 << _bitDepth) - 1;

    for (int y = 0; y < _height; ++y) {
        uint16_t* row = _imageData.ptr<uint16_t>(y);
        for (int x = 0; x < _width; ++x) {
            row[x] = extractBits(rawData, bitPos, _bitDepth, maxValue);
            bitPos += _bitDepth;
        }
    }
}

uint16_t DXF::extractBits(const vector<uint8_t>& buffer, size_t bitPos, int bitDepth, uint16_t maxValue) {
    size_t bytePos = bitPos / 8;
    size_t bitOffset = bitPos % 8;

    uint32_t temp = 0;
    for (int i = 0; i < 3 && bytePos + i < buffer.size(); ++i) {
        temp |= static_cast<uint32_t>(buffer[bytePos + i]) << (16 - (i * 8));
    }

    temp >>= (24 - (bitOffset + bitDepth));
    return static_cast<uint16_t>(temp & maxValue);
}

void DXF::calculateDimensions(streamsize fileSize) {
    int pixelCount = (fileSize * 8) / _bitDepth;
    _width = static_cast<int>(sqrt(pixelCount));
    _height = _width;
}

void DXF::validateImageProperties() {
    if (_width <= 0 || _height <= 0 || 
        _width > 65536 || _height > 65536) {
        throw runtime_error("Invalid image dimensions");
    }

    if (_bitDepth < 8 || _bitDepth > 16) {
        throw runtime_error("Unsupported bit depth: " + to_string(_bitDepth));
    }
}